"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const request = require("superagent");
const handleError = (err) => {
    if (err.response && err.response.body) {
        const body = err.response.body;
        const url = err.response.request.url;
        const method = err.response.request.method.toUpperCase();
        const status = err.response.status;
        if (body.error && body.error.message) {
            throw new Error(`Cannot ${method} ${url} (${status}): ${body.error.message}`);
        }
    }
    throw err;
};
const handleResponse = (resp, timer) => {
    return {
        result: resp.body,
        properties: {
            requestId: resp.header['request-id'],
            elapsedTime: parseInt(resp.header['elapsed-time'], 10),
            clientRequestId: resp.header['client-request-id'],
            eTag: resp.header.etag,
            location: resp.header.location,
        },
        statusCode: resp.status,
        timer,
    };
};
const handlePromise = (req, timer) => {
    return req
        .then((resp) => handleResponse(resp, timer))
        .catch(handleError);
};
const handleCallback = (req, callback, timer) => {
    req.end((err, resp) => {
        let error;
        let searchResp;
        if (err) {
            try {
                handleError(err);
            }
            catch (err) {
                error = err;
            }
        }
        else {
            searchResp = handleResponse(resp, timer);
        }
        callback(error, searchResp);
    });
};
request.serialize['application/json'] = (obj) => {
    return Buffer.isBuffer(obj) ? obj : JSON.stringify(obj);
};
/** Internal class to handle search HTTP requests/responses */
class SearchRequester {
    constructor(service, adminKey, defaultVersion) {
        this.service = service;
        this.defaultVersion = defaultVersion;
        this.events = new events_1.EventEmitter();
        this.adminKeys = Array.isArray(adminKey) ? adminKey : [adminKey];
    }
    get endpoint() { return `https://${this.service}.search.windows.net`; }
    request(req, optionsOrCallback, callback) {
        const [options, cb] = this.getParams(optionsOrCallback, callback);
        const events = this.events;
        const headers = Object.assign({
            'api-key': options && options.key ? options.key : this.adminKeys[0],
            'if-match': options && options.ifMatch ? options.ifMatch : null,
            'if-none-match': options && options.ifNoneMatch ? options.ifNoneMatch : null,
            'client-request-id': options && options.clientRequestId ? options.clientRequestId : null,
            'return-client-request-id': options && options.returnClientRequestId ? 'True' : null,
        }, req.headers);
        const query = Object.assign({
            'api-version': options && options.version ? options.version : this.defaultVersion,
        }, req.query);
        const timer = { start: new Date(), response: process.hrtime(), end: process.hrtime() };
        const val = request(req.method, this.endpoint + req.path)
            .set(headers)
            .query(query)
            .send(req.body)
            .retry(options ? options.retry : 0)
            .timeout(options ? options.timeout : null)
            .parse(req.parser)
            .on('response', (resp) => {
            timer.response = process.hrtime(timer.response);
            this.events.emit('response', resp);
        })
            .on('error', (err) => {
            timer.end = process.hrtime(timer.end);
            if (this.events.listenerCount('error')) {
                this.events.emit('error', err);
            }
        })
            .on('end', (err) => {
            timer.end = process.hrtime(timer.end);
        });
        this.events.emit('request', { request: req, options });
        if (cb) {
            handleCallback(val, cb, timer);
        }
        else {
            return handlePromise(val, timer);
        }
    }
    getParams(optionsOrCallback, callback) {
        if (typeof optionsOrCallback === 'function') {
            return [{}, optionsOrCallback];
        }
        else {
            return [optionsOrCallback || {}, callback];
        }
    }
}
exports.SearchRequester = SearchRequester;
//# sourceMappingURL=search-requester.js.map